/* 
 * Copyright (C) 2013-2014 Pier Luigi Fiorini
 * 
 * Permission to use, copy, modify, distribute, and sell this
 * software and its documentation for any purpose is hereby granted
 * without fee, provided that the above copyright notice appear in
 * all copies and that both that copyright notice and this permission
 * notice appear in supporting documentation, and that the name of
 * the copyright holders not be used in advertising or publicity
 * pertaining to distribution of the software without specific,
 * written prior permission.  The copyright holders make no
 * representations about the suitability of this software for any
 * purpose.  It is provided "as is" without express or implied
 * warranty.
 * 
 * THE COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS, IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
 * AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
 * THIS SOFTWARE.
 */

#ifndef HAWAII_CLIENT_PROTOCOL_H
#define HAWAII_CLIENT_PROTOCOL_H

#ifdef  __cplusplus
extern "C" {
#endif

#include <stdint.h>
#include <stddef.h>
#include "wayland-client.h"

struct wl_client;
struct wl_resource;

struct hawaii_shell;
struct hawaii_panel_manager;
struct hawaii_panel;
struct hawaii_shell_surface;
struct hawaii_popup_surface;
struct hawaii_key_binding;
struct hawaii_window;
struct hawaii_workspace;

extern const struct wl_interface hawaii_shell_interface;
extern const struct wl_interface hawaii_panel_manager_interface;
extern const struct wl_interface hawaii_panel_interface;
extern const struct wl_interface hawaii_shell_surface_interface;
extern const struct wl_interface hawaii_popup_surface_interface;
extern const struct wl_interface hawaii_key_binding_interface;
extern const struct wl_interface hawaii_window_interface;
extern const struct wl_interface hawaii_workspace_interface;

#ifndef HAWAII_SHELL_CURSOR_ENUM
#define HAWAII_SHELL_CURSOR_ENUM
enum hawaii_shell_cursor {
	HAWAII_SHELL_CURSOR_NONE = 0,
	HAWAII_SHELL_CURSOR_RESIZE_TOP = 1,
	HAWAII_SHELL_CURSOR_RESIZE_BOTTOM = 2,
	HAWAII_SHELL_CURSOR_ARROW = 3,
	HAWAII_SHELL_CURSOR_RESIZE_LEFT = 4,
	HAWAII_SHELL_CURSOR_RESIZE_TOP_LEFT = 5,
	HAWAII_SHELL_CURSOR_RESIZE_BOTTOM_LEFT = 6,
	HAWAII_SHELL_CURSOR_MOVE = 7,
	HAWAII_SHELL_CURSOR_RESIZE_RIGHT = 8,
	HAWAII_SHELL_CURSOR_RESIZE_TOP_RIGHT = 9,
	HAWAII_SHELL_CURSOR_RESIZE_BOTTOM_RIGHT = 10,
	HAWAII_SHELL_CURSOR_BUSY = 11,
};
#endif /* HAWAII_SHELL_CURSOR_ENUM */

#ifndef HAWAII_SHELL_WINDOW_STATE_ENUM
#define HAWAII_SHELL_WINDOW_STATE_ENUM
enum hawaii_shell_window_state {
	HAWAII_SHELL_WINDOW_STATE_INACTIVE = 0,
	HAWAII_SHELL_WINDOW_STATE_ACTIVE = 1,
	HAWAII_SHELL_WINDOW_STATE_MINIMIZED = 2,
	HAWAII_SHELL_WINDOW_STATE_MAXIMIZED = 4,
	HAWAII_SHELL_WINDOW_STATE_FULLSCREEN = 8,
};
#endif /* HAWAII_SHELL_WINDOW_STATE_ENUM */

/**
 * hawaii_shell - create desktop widgets and helpers
 * @loaded: tell the client to create and set its surfaces
 * @configure: suggest resize
 * @prepare_lock_surface: tell the client to create, set the lock surface
 * @grab_cursor: tell client what cursor to show during a grab
 * @window_mapped: notify the client that a new window was created
 * @window_switching_started: tell client to show the window switching
 *	overlay
 * @window_switching_finished: tell client to hide the window switching
 *	overlay
 * @window_switched: notify the client that the server switched to a
 *	window
 * @workspace_added: notify the client that a workspace was added
 *
 * This interface defines the foundations of the Hawaii desktop.
 */
struct hawaii_shell_listener {
	/**
	 * loaded - tell the client to create and set its surfaces
	 *
	 * Tell the shell we want it to create and set its surfaces such
	 * as background and panels. The client will emit desktop_ready
	 * when all surfaces are ready.
	 */
	void (*loaded)(void *data,
		       struct hawaii_shell *hawaii_shell);
	/**
	 * configure - suggest resize
	 * @surface: (none)
	 * @width: (none)
	 * @height: (none)
	 *
	 * The configure event asks the shell to resize its surface.
	 *
	 * The width and height arguments specify the size of the window in
	 * surface local coordinates.
	 */
	void (*configure)(void *data,
			  struct hawaii_shell *hawaii_shell,
			  struct wl_surface *surface,
			  int32_t width,
			  int32_t height);
	/**
	 * prepare_lock_surface - tell the client to create, set the lock
	 *	surface
	 *
	 * Tell the shell we want it to create and set the lock surface,
	 * which is a GUI asking the user to unlock the screen. The lock
	 * surface is announced with 'set_lock_surface'. Whether or not the
	 * shell actually implements locking, it MUST send 'unlock' request
	 * to let the normal desktop resume.
	 */
	void (*prepare_lock_surface)(void *data,
				     struct hawaii_shell *hawaii_shell);
	/**
	 * grab_cursor - tell client what cursor to show during a grab
	 * @cursor: (none)
	 *
	 * This event will be sent immediately before a fake enter event
	 * on the grab surface.
	 */
	void (*grab_cursor)(void *data,
			    struct hawaii_shell *hawaii_shell,
			    uint32_t cursor);
	/**
	 * window_mapped - notify the client that a new window was
	 *	created
	 * @id: (none)
	 * @title: (none)
	 * @identifier: (none)
	 * @state: (none)
	 *
	 * This event will be sent immediately after a new window is
	 * mapped.
	 */
	void (*window_mapped)(void *data,
			      struct hawaii_shell *hawaii_shell,
			      struct hawaii_window *id,
			      const char *title,
			      const char *identifier,
			      int32_t state);
	/**
	 * window_switching_started - tell client to show the window
	 *	switching overlay
	 *
	 * This event will be sent immediately after window switching is
	 * started.
	 */
	void (*window_switching_started)(void *data,
					 struct hawaii_shell *hawaii_shell);
	/**
	 * window_switching_finished - tell client to hide the window
	 *	switching overlay
	 *
	 * This event will be sent immediately after window switching
	 * finished.
	 */
	void (*window_switching_finished)(void *data,
					  struct hawaii_shell *hawaii_shell);
	/**
	 * window_switched - notify the client that the server switched
	 *	to a window
	 * @window: (none)
	 *
	 * Windows switching is handled by the server. The client
	 * provides a nice user interface that shows which window the
	 * server switched to. This event notifies the client which window
	 * the server switched to.
	 */
	void (*window_switched)(void *data,
				struct hawaii_shell *hawaii_shell,
				struct hawaii_window *window);
	/**
	 * workspace_added - notify the client that a workspace was added
	 * @workspace: (none)
	 * @active: (none)
	 *
	 * This event will be sent immediately after a workspace is
	 * added.
	 */
	void (*workspace_added)(void *data,
				struct hawaii_shell *hawaii_shell,
				struct hawaii_workspace *workspace,
				int32_t active);
};

static inline int
hawaii_shell_add_listener(struct hawaii_shell *hawaii_shell,
			  const struct hawaii_shell_listener *listener, void *data)
{
	return wl_proxy_add_listener((struct wl_proxy *) hawaii_shell,
				     (void (**)(void)) listener, data);
}

#define HAWAII_SHELL_ADD_TRUSTED_CLIENT	0
#define HAWAII_SHELL_ADD_KEY_BINDING	1
#define HAWAII_SHELL_SET_BACKGROUND	2
#define HAWAII_SHELL_SET_DESKTOP	3
#define HAWAII_SHELL_SET_CONFIG_SURFACE	4
#define HAWAII_SHELL_SET_OVERLAY	5
#define HAWAII_SHELL_SET_LOCK_SURFACE	6
#define HAWAII_SHELL_SET_GRAB_SURFACE	7
#define HAWAII_SHELL_SET_POSITION	8
#define HAWAII_SHELL_QUIT	9
#define HAWAII_SHELL_LOCK	10
#define HAWAII_SHELL_UNLOCK	11
#define HAWAII_SHELL_DESKTOP_READY	12
#define HAWAII_SHELL_MINIMIZE_WINDOWS	13
#define HAWAII_SHELL_RESTORE_WINDOWS	14
#define HAWAII_SHELL_ADD_WORKSPACE	15
#define HAWAII_SHELL_SELECT_WORKSPACE	16

static inline void
hawaii_shell_set_user_data(struct hawaii_shell *hawaii_shell, void *user_data)
{
	wl_proxy_set_user_data((struct wl_proxy *) hawaii_shell, user_data);
}

static inline void *
hawaii_shell_get_user_data(struct hawaii_shell *hawaii_shell)
{
	return wl_proxy_get_user_data((struct wl_proxy *) hawaii_shell);
}

static inline void
hawaii_shell_destroy(struct hawaii_shell *hawaii_shell)
{
	wl_proxy_destroy((struct wl_proxy *) hawaii_shell);
}

static inline void
hawaii_shell_add_trusted_client(struct hawaii_shell *hawaii_shell, int32_t fd, const char *interface)
{
	wl_proxy_marshal((struct wl_proxy *) hawaii_shell,
			 HAWAII_SHELL_ADD_TRUSTED_CLIENT, fd, interface);
}

static inline struct hawaii_key_binding *
hawaii_shell_add_key_binding(struct hawaii_shell *hawaii_shell, uint32_t key, uint32_t modifiers)
{
	struct wl_proxy *id;

	id = wl_proxy_marshal_constructor((struct wl_proxy *) hawaii_shell,
			 HAWAII_SHELL_ADD_KEY_BINDING, &hawaii_key_binding_interface, NULL, key, modifiers);

	return (struct hawaii_key_binding *) id;
}

static inline void
hawaii_shell_set_background(struct hawaii_shell *hawaii_shell, struct wl_output *output, struct wl_surface *surface)
{
	wl_proxy_marshal((struct wl_proxy *) hawaii_shell,
			 HAWAII_SHELL_SET_BACKGROUND, output, surface);
}

static inline void
hawaii_shell_set_desktop(struct hawaii_shell *hawaii_shell, struct wl_output *output, struct wl_surface *surface)
{
	wl_proxy_marshal((struct wl_proxy *) hawaii_shell,
			 HAWAII_SHELL_SET_DESKTOP, output, surface);
}

static inline void
hawaii_shell_set_config_surface(struct hawaii_shell *hawaii_shell, struct wl_output *output, struct wl_surface *surface)
{
	wl_proxy_marshal((struct wl_proxy *) hawaii_shell,
			 HAWAII_SHELL_SET_CONFIG_SURFACE, output, surface);
}

static inline void
hawaii_shell_set_overlay(struct hawaii_shell *hawaii_shell, struct wl_surface *surface)
{
	wl_proxy_marshal((struct wl_proxy *) hawaii_shell,
			 HAWAII_SHELL_SET_OVERLAY, surface);
}

static inline void
hawaii_shell_set_lock_surface(struct hawaii_shell *hawaii_shell, struct wl_surface *surface)
{
	wl_proxy_marshal((struct wl_proxy *) hawaii_shell,
			 HAWAII_SHELL_SET_LOCK_SURFACE, surface);
}

static inline void
hawaii_shell_set_grab_surface(struct hawaii_shell *hawaii_shell, struct wl_surface *surface)
{
	wl_proxy_marshal((struct wl_proxy *) hawaii_shell,
			 HAWAII_SHELL_SET_GRAB_SURFACE, surface);
}

static inline void
hawaii_shell_set_position(struct hawaii_shell *hawaii_shell, struct wl_surface *surface, int32_t x, int32_t y)
{
	wl_proxy_marshal((struct wl_proxy *) hawaii_shell,
			 HAWAII_SHELL_SET_POSITION, surface, x, y);
}

static inline void
hawaii_shell_quit(struct hawaii_shell *hawaii_shell)
{
	wl_proxy_marshal((struct wl_proxy *) hawaii_shell,
			 HAWAII_SHELL_QUIT);
}

static inline void
hawaii_shell_lock(struct hawaii_shell *hawaii_shell)
{
	wl_proxy_marshal((struct wl_proxy *) hawaii_shell,
			 HAWAII_SHELL_LOCK);
}

static inline void
hawaii_shell_unlock(struct hawaii_shell *hawaii_shell)
{
	wl_proxy_marshal((struct wl_proxy *) hawaii_shell,
			 HAWAII_SHELL_UNLOCK);
}

static inline void
hawaii_shell_desktop_ready(struct hawaii_shell *hawaii_shell)
{
	wl_proxy_marshal((struct wl_proxy *) hawaii_shell,
			 HAWAII_SHELL_DESKTOP_READY);
}

static inline void
hawaii_shell_minimize_windows(struct hawaii_shell *hawaii_shell)
{
	wl_proxy_marshal((struct wl_proxy *) hawaii_shell,
			 HAWAII_SHELL_MINIMIZE_WINDOWS);
}

static inline void
hawaii_shell_restore_windows(struct hawaii_shell *hawaii_shell)
{
	wl_proxy_marshal((struct wl_proxy *) hawaii_shell,
			 HAWAII_SHELL_RESTORE_WINDOWS);
}

static inline void
hawaii_shell_add_workspace(struct hawaii_shell *hawaii_shell)
{
	wl_proxy_marshal((struct wl_proxy *) hawaii_shell,
			 HAWAII_SHELL_ADD_WORKSPACE);
}

static inline void
hawaii_shell_select_workspace(struct hawaii_shell *hawaii_shell, struct hawaii_workspace *workspace)
{
	wl_proxy_marshal((struct wl_proxy *) hawaii_shell,
			 HAWAII_SHELL_SELECT_WORKSPACE, workspace);
}

#define HAWAII_PANEL_MANAGER_SET_PANEL	0

static inline void
hawaii_panel_manager_set_user_data(struct hawaii_panel_manager *hawaii_panel_manager, void *user_data)
{
	wl_proxy_set_user_data((struct wl_proxy *) hawaii_panel_manager, user_data);
}

static inline void *
hawaii_panel_manager_get_user_data(struct hawaii_panel_manager *hawaii_panel_manager)
{
	return wl_proxy_get_user_data((struct wl_proxy *) hawaii_panel_manager);
}

static inline void
hawaii_panel_manager_destroy(struct hawaii_panel_manager *hawaii_panel_manager)
{
	wl_proxy_destroy((struct wl_proxy *) hawaii_panel_manager);
}

static inline struct hawaii_panel *
hawaii_panel_manager_set_panel(struct hawaii_panel_manager *hawaii_panel_manager, struct wl_surface *surface)
{
	struct wl_proxy *id;

	id = wl_proxy_marshal_constructor((struct wl_proxy *) hawaii_panel_manager,
			 HAWAII_PANEL_MANAGER_SET_PANEL, &hawaii_panel_interface, NULL, surface);

	return (struct hawaii_panel *) id;
}

#ifndef HAWAII_PANEL_EDGE_ENUM
#define HAWAII_PANEL_EDGE_ENUM
enum hawaii_panel_edge {
	HAWAII_PANEL_EDGE_LEFT = 0,
	HAWAII_PANEL_EDGE_TOP = 1,
	HAWAII_PANEL_EDGE_RIGHT = 2,
	HAWAII_PANEL_EDGE_BOTTOM = 3,
};
#endif /* HAWAII_PANEL_EDGE_ENUM */

#ifndef HAWAII_PANEL_ALIGNMENT_ENUM
#define HAWAII_PANEL_ALIGNMENT_ENUM
enum hawaii_panel_alignment {
	HAWAII_PANEL_ALIGNMENT_LEFT = 0,
	HAWAII_PANEL_ALIGNMENT_CENTER = 1,
	HAWAII_PANEL_ALIGNMENT_RIGHT = 2,
};
#endif /* HAWAII_PANEL_ALIGNMENT_ENUM */

struct hawaii_panel_listener {
	/**
	 * docked - panel is docked
	 *
	 * Emitted when the panel is docked.
	 */
	void (*docked)(void *data,
		       struct hawaii_panel *hawaii_panel);
};

static inline int
hawaii_panel_add_listener(struct hawaii_panel *hawaii_panel,
			  const struct hawaii_panel_listener *listener, void *data)
{
	return wl_proxy_add_listener((struct wl_proxy *) hawaii_panel,
				     (void (**)(void)) listener, data);
}

#define HAWAII_PANEL_SET_ALIGNMENT	0
#define HAWAII_PANEL_SET_OFFSET	1
#define HAWAII_PANEL_SET_THICKNESS	2
#define HAWAII_PANEL_SET_LENGTH	3
#define HAWAII_PANEL_SET_MIN_LENGTH	4
#define HAWAII_PANEL_SET_MAX_LENGTH	5
#define HAWAII_PANEL_DOCK	6

static inline void
hawaii_panel_set_user_data(struct hawaii_panel *hawaii_panel, void *user_data)
{
	wl_proxy_set_user_data((struct wl_proxy *) hawaii_panel, user_data);
}

static inline void *
hawaii_panel_get_user_data(struct hawaii_panel *hawaii_panel)
{
	return wl_proxy_get_user_data((struct wl_proxy *) hawaii_panel);
}

static inline void
hawaii_panel_destroy(struct hawaii_panel *hawaii_panel)
{
	wl_proxy_destroy((struct wl_proxy *) hawaii_panel);
}

static inline void
hawaii_panel_set_alignment(struct hawaii_panel *hawaii_panel, uint32_t alignment)
{
	wl_proxy_marshal((struct wl_proxy *) hawaii_panel,
			 HAWAII_PANEL_SET_ALIGNMENT, alignment);
}

static inline void
hawaii_panel_set_offset(struct hawaii_panel *hawaii_panel, uint32_t offset)
{
	wl_proxy_marshal((struct wl_proxy *) hawaii_panel,
			 HAWAII_PANEL_SET_OFFSET, offset);
}

static inline void
hawaii_panel_set_thickness(struct hawaii_panel *hawaii_panel, uint32_t thickness)
{
	wl_proxy_marshal((struct wl_proxy *) hawaii_panel,
			 HAWAII_PANEL_SET_THICKNESS, thickness);
}

static inline void
hawaii_panel_set_length(struct hawaii_panel *hawaii_panel, uint32_t length)
{
	wl_proxy_marshal((struct wl_proxy *) hawaii_panel,
			 HAWAII_PANEL_SET_LENGTH, length);
}

static inline void
hawaii_panel_set_min_length(struct hawaii_panel *hawaii_panel, uint32_t min_length)
{
	wl_proxy_marshal((struct wl_proxy *) hawaii_panel,
			 HAWAII_PANEL_SET_MIN_LENGTH, min_length);
}

static inline void
hawaii_panel_set_max_length(struct hawaii_panel *hawaii_panel, uint32_t max_length)
{
	wl_proxy_marshal((struct wl_proxy *) hawaii_panel,
			 HAWAII_PANEL_SET_MAX_LENGTH, max_length);
}

static inline void
hawaii_panel_dock(struct hawaii_panel *hawaii_panel, uint32_t edge, struct wl_output *output)
{
	wl_proxy_marshal((struct wl_proxy *) hawaii_panel,
			 HAWAII_PANEL_DOCK, edge, output);
}

#define HAWAII_SHELL_SURFACE_SET_POPUP	0
#define HAWAII_SHELL_SURFACE_SET_DIALOG	1

static inline void
hawaii_shell_surface_set_user_data(struct hawaii_shell_surface *hawaii_shell_surface, void *user_data)
{
	wl_proxy_set_user_data((struct wl_proxy *) hawaii_shell_surface, user_data);
}

static inline void *
hawaii_shell_surface_get_user_data(struct hawaii_shell_surface *hawaii_shell_surface)
{
	return wl_proxy_get_user_data((struct wl_proxy *) hawaii_shell_surface);
}

static inline void
hawaii_shell_surface_destroy(struct hawaii_shell_surface *hawaii_shell_surface)
{
	wl_proxy_destroy((struct wl_proxy *) hawaii_shell_surface);
}

static inline struct hawaii_popup_surface *
hawaii_shell_surface_set_popup(struct hawaii_shell_surface *hawaii_shell_surface, struct wl_surface *parent, struct wl_surface *surface, int32_t x, int32_t y)
{
	struct wl_proxy *id;

	id = wl_proxy_marshal_constructor((struct wl_proxy *) hawaii_shell_surface,
			 HAWAII_SHELL_SURFACE_SET_POPUP, &hawaii_popup_surface_interface, NULL, parent, surface, x, y);

	return (struct hawaii_popup_surface *) id;
}

static inline void
hawaii_shell_surface_set_dialog(struct hawaii_shell_surface *hawaii_shell_surface, struct wl_output *output, struct wl_surface *surface)
{
	wl_proxy_marshal((struct wl_proxy *) hawaii_shell_surface,
			 HAWAII_SHELL_SURFACE_SET_DIALOG, output, surface);
}

struct hawaii_popup_surface_listener {
	/**
	 * popup_done - popup interaction is done
	 *
	 * The popup_done event is sent out when a popup grab is broken,
	 * that is, when the users clicks on a different surface.
	 */
	void (*popup_done)(void *data,
			   struct hawaii_popup_surface *hawaii_popup_surface);
};

static inline int
hawaii_popup_surface_add_listener(struct hawaii_popup_surface *hawaii_popup_surface,
				  const struct hawaii_popup_surface_listener *listener, void *data)
{
	return wl_proxy_add_listener((struct wl_proxy *) hawaii_popup_surface,
				     (void (**)(void)) listener, data);
}

static inline void
hawaii_popup_surface_set_user_data(struct hawaii_popup_surface *hawaii_popup_surface, void *user_data)
{
	wl_proxy_set_user_data((struct wl_proxy *) hawaii_popup_surface, user_data);
}

static inline void *
hawaii_popup_surface_get_user_data(struct hawaii_popup_surface *hawaii_popup_surface)
{
	return wl_proxy_get_user_data((struct wl_proxy *) hawaii_popup_surface);
}

static inline void
hawaii_popup_surface_destroy(struct hawaii_popup_surface *hawaii_popup_surface)
{
	wl_proxy_destroy((struct wl_proxy *) hawaii_popup_surface);
}

/**
 * hawaii_key_binding - interface for key bindings
 * @triggered: key binding was triggered
 *
 * This interface represents a key binding.
 */
struct hawaii_key_binding_listener {
	/**
	 * triggered - key binding was triggered
	 *
	 * This event will be sent when a registered key binding will be
	 * triggered.
	 */
	void (*triggered)(void *data,
			  struct hawaii_key_binding *hawaii_key_binding);
};

static inline int
hawaii_key_binding_add_listener(struct hawaii_key_binding *hawaii_key_binding,
				const struct hawaii_key_binding_listener *listener, void *data)
{
	return wl_proxy_add_listener((struct wl_proxy *) hawaii_key_binding,
				     (void (**)(void)) listener, data);
}

static inline void
hawaii_key_binding_set_user_data(struct hawaii_key_binding *hawaii_key_binding, void *user_data)
{
	wl_proxy_set_user_data((struct wl_proxy *) hawaii_key_binding, user_data);
}

static inline void *
hawaii_key_binding_get_user_data(struct hawaii_key_binding *hawaii_key_binding)
{
	return wl_proxy_get_user_data((struct wl_proxy *) hawaii_key_binding);
}

static inline void
hawaii_key_binding_destroy(struct hawaii_key_binding *hawaii_key_binding)
{
	wl_proxy_destroy((struct wl_proxy *) hawaii_key_binding);
}

/**
 * hawaii_window - interface to control client windows
 * @title_changed: window title have been changed
 * @identifier_changed: window class have been changed
 * @state_changed: window state have been changed
 * @unmapped: window's surface was unmapped
 *
 * Only one client can bind this interface at a time.
 */
struct hawaii_window_listener {
	/**
	 * title_changed - window title have been changed
	 * @title: (none)
	 *
	 * This event will be sent as soon as the window title is
	 * changed.
	 */
	void (*title_changed)(void *data,
			      struct hawaii_window *hawaii_window,
			      const char *title);
	/**
	 * identifier_changed - window class have been changed
	 * @identifier: (none)
	 *
	 * This event will be sent as soon as the window class is
	 * changed.
	 */
	void (*identifier_changed)(void *data,
				   struct hawaii_window *hawaii_window,
				   const char *identifier);
	/**
	 * state_changed - window state have been changed
	 * @state: (none)
	 *
	 * This event will be sent as soon as the window state changes.
	 */
	void (*state_changed)(void *data,
			      struct hawaii_window *hawaii_window,
			      int32_t state);
	/**
	 * unmapped - window's surface was unmapped
	 *
	 * This event will be sent immediately after the window is closed
	 * and its surface is unmapped.
	 */
	void (*unmapped)(void *data,
			 struct hawaii_window *hawaii_window);
};

static inline int
hawaii_window_add_listener(struct hawaii_window *hawaii_window,
			   const struct hawaii_window_listener *listener, void *data)
{
	return wl_proxy_add_listener((struct wl_proxy *) hawaii_window,
				     (void (**)(void)) listener, data);
}

#define HAWAII_WINDOW_SET_STATE	0
#define HAWAII_WINDOW_CLOSE	1

static inline void
hawaii_window_set_user_data(struct hawaii_window *hawaii_window, void *user_data)
{
	wl_proxy_set_user_data((struct wl_proxy *) hawaii_window, user_data);
}

static inline void *
hawaii_window_get_user_data(struct hawaii_window *hawaii_window)
{
	return wl_proxy_get_user_data((struct wl_proxy *) hawaii_window);
}

static inline void
hawaii_window_destroy(struct hawaii_window *hawaii_window)
{
	wl_proxy_destroy((struct wl_proxy *) hawaii_window);
}

static inline void
hawaii_window_set_state(struct hawaii_window *hawaii_window, int32_t state)
{
	wl_proxy_marshal((struct wl_proxy *) hawaii_window,
			 HAWAII_WINDOW_SET_STATE, state);
}

static inline void
hawaii_window_close(struct hawaii_window *hawaii_window)
{
	wl_proxy_marshal((struct wl_proxy *) hawaii_window,
			 HAWAII_WINDOW_CLOSE);
}

/**
 * hawaii_workspace - control and manager workspaces
 * @activated: workspace activated
 * @deactivated: workspace deactivated
 *
 * Each workspace can bind this interface.
 */
struct hawaii_workspace_listener {
	/**
	 * activated - workspace activated
	 *
	 * This event will be sent immediately after this workspace is
	 * activated.
	 */
	void (*activated)(void *data,
			  struct hawaii_workspace *hawaii_workspace);
	/**
	 * deactivated - workspace deactivated
	 *
	 * This event will be sent immediately after this workspace is
	 * deactivated.
	 */
	void (*deactivated)(void *data,
			    struct hawaii_workspace *hawaii_workspace);
};

static inline int
hawaii_workspace_add_listener(struct hawaii_workspace *hawaii_workspace,
			      const struct hawaii_workspace_listener *listener, void *data)
{
	return wl_proxy_add_listener((struct wl_proxy *) hawaii_workspace,
				     (void (**)(void)) listener, data);
}

#define HAWAII_WORKSPACE_REMOVE	0

static inline void
hawaii_workspace_set_user_data(struct hawaii_workspace *hawaii_workspace, void *user_data)
{
	wl_proxy_set_user_data((struct wl_proxy *) hawaii_workspace, user_data);
}

static inline void *
hawaii_workspace_get_user_data(struct hawaii_workspace *hawaii_workspace)
{
	return wl_proxy_get_user_data((struct wl_proxy *) hawaii_workspace);
}

static inline void
hawaii_workspace_destroy(struct hawaii_workspace *hawaii_workspace)
{
	wl_proxy_destroy((struct wl_proxy *) hawaii_workspace);
}

static inline void
hawaii_workspace_remove(struct hawaii_workspace *hawaii_workspace)
{
	wl_proxy_marshal((struct wl_proxy *) hawaii_workspace,
			 HAWAII_WORKSPACE_REMOVE);
}

#ifdef  __cplusplus
}
#endif

#endif
