/* 
 * Copyright (C) 2013-2014 Pier Luigi Fiorini
 * 
 * Permission to use, copy, modify, distribute, and sell this
 * software and its documentation for any purpose is hereby granted
 * without fee, provided that the above copyright notice appear in
 * all copies and that both that copyright notice and this permission
 * notice appear in supporting documentation, and that the name of
 * the copyright holders not be used in advertising or publicity
 * pertaining to distribution of the software without specific,
 * written prior permission.  The copyright holders make no
 * representations about the suitability of this software for any
 * purpose.  It is provided "as is" without express or implied
 * warranty.
 * 
 * THE COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS, IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
 * AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
 * THIS SOFTWARE.
 */

#ifndef HAWAII_SERVER_PROTOCOL_H
#define HAWAII_SERVER_PROTOCOL_H

#ifdef  __cplusplus
extern "C" {
#endif

#include <stdint.h>
#include <stddef.h>
#include "wayland-util.h"

struct wl_client;
struct wl_resource;

struct hawaii_shell;
struct hawaii_panel_manager;
struct hawaii_panel;
struct hawaii_shell_surface;
struct hawaii_popup_surface;
struct hawaii_key_binding;
struct hawaii_window;
struct hawaii_workspace;

extern const struct wl_interface hawaii_shell_interface;
extern const struct wl_interface hawaii_panel_manager_interface;
extern const struct wl_interface hawaii_panel_interface;
extern const struct wl_interface hawaii_shell_surface_interface;
extern const struct wl_interface hawaii_popup_surface_interface;
extern const struct wl_interface hawaii_key_binding_interface;
extern const struct wl_interface hawaii_window_interface;
extern const struct wl_interface hawaii_workspace_interface;

#ifndef HAWAII_SHELL_CURSOR_ENUM
#define HAWAII_SHELL_CURSOR_ENUM
enum hawaii_shell_cursor {
	HAWAII_SHELL_CURSOR_NONE = 0,
	HAWAII_SHELL_CURSOR_RESIZE_TOP = 1,
	HAWAII_SHELL_CURSOR_RESIZE_BOTTOM = 2,
	HAWAII_SHELL_CURSOR_ARROW = 3,
	HAWAII_SHELL_CURSOR_RESIZE_LEFT = 4,
	HAWAII_SHELL_CURSOR_RESIZE_TOP_LEFT = 5,
	HAWAII_SHELL_CURSOR_RESIZE_BOTTOM_LEFT = 6,
	HAWAII_SHELL_CURSOR_MOVE = 7,
	HAWAII_SHELL_CURSOR_RESIZE_RIGHT = 8,
	HAWAII_SHELL_CURSOR_RESIZE_TOP_RIGHT = 9,
	HAWAII_SHELL_CURSOR_RESIZE_BOTTOM_RIGHT = 10,
	HAWAII_SHELL_CURSOR_BUSY = 11,
};
#endif /* HAWAII_SHELL_CURSOR_ENUM */

#ifndef HAWAII_SHELL_WINDOW_STATE_ENUM
#define HAWAII_SHELL_WINDOW_STATE_ENUM
enum hawaii_shell_window_state {
	HAWAII_SHELL_WINDOW_STATE_INACTIVE = 0,
	HAWAII_SHELL_WINDOW_STATE_ACTIVE = 1,
	HAWAII_SHELL_WINDOW_STATE_MINIMIZED = 2,
	HAWAII_SHELL_WINDOW_STATE_MAXIMIZED = 4,
	HAWAII_SHELL_WINDOW_STATE_FULLSCREEN = 8,
};
#endif /* HAWAII_SHELL_WINDOW_STATE_ENUM */

/**
 * hawaii_shell - create desktop widgets and helpers
 * @add_trusted_client: add a trusted client
 * @add_key_binding: add key binding
 * @set_background: set background role
 * @set_desktop: set desktop role
 * @set_config_surface: set configuration role
 * @set_overlay: set overlay role
 * @set_lock_surface: set lock role
 * @set_grab_surface: set grab surface
 * @set_position: change surface position
 * @quit: quit the compositor
 * @lock: lock the session
 * @unlock: unlock the session
 * @desktop_ready: desktop is ready to be shown
 * @minimize_windows: minimize all windows
 * @restore_windows: restore all windows
 * @add_workspace: add workspace
 * @select_workspace: select workspace
 *
 * This interface defines the foundations of the Hawaii desktop.
 */
struct hawaii_shell_interface {
	/**
	 * add_trusted_client - add a trusted client
	 * @fd: (none)
	 * @interface: (none)
	 *
	 * Adds a client to the whitelist for the specified interface.
	 */
	void (*add_trusted_client)(struct wl_client *client,
				   struct wl_resource *resource,
				   int32_t fd,
				   const char *interface);
	/**
	 * add_key_binding - add key binding
	 * @id: (none)
	 * @key: (none)
	 * @modifiers: (none)
	 *
	 * Bind a global shortcut.
	 */
	void (*add_key_binding)(struct wl_client *client,
				struct wl_resource *resource,
				uint32_t id,
				uint32_t key,
				uint32_t modifiers);
	/**
	 * set_background - set background role
	 * @output: (none)
	 * @surface: (none)
	 *
	 * The surface set by this request will receive the background
	 * role. Background surfaces are placed above all surfaces and can
	 * be used to show wallpapers.
	 */
	void (*set_background)(struct wl_client *client,
			       struct wl_resource *resource,
			       struct wl_resource *output,
			       struct wl_resource *surface);
	/**
	 * set_desktop - set desktop role
	 * @output: (none)
	 * @surface: (none)
	 *
	 * The surface set by this request will receive the desktop role.
	 * Desktop surfaces are placed above background surfaces and can be
	 * used to show the actual desktop view with icons, search results
	 * or controls the user will interact with.
	 */
	void (*set_desktop)(struct wl_client *client,
			    struct wl_resource *resource,
			    struct wl_resource *output,
			    struct wl_resource *surface);
	/**
	 * set_config_surface - set configuration role
	 * @output: (none)
	 * @surface: (none)
	 *
	 * The surface set by this request will receive the configuration
	 * role. A configuration surface is shown when the user wants to
	 * configure panel or desktop views.
	 */
	void (*set_config_surface)(struct wl_client *client,
				   struct wl_resource *resource,
				   struct wl_resource *output,
				   struct wl_resource *surface);
	/**
	 * set_overlay - set overlay role
	 * @surface: (none)
	 *
	 * The surface set by this request will receive the overlay role.
	 * Overlays are special surfaces that shows for a limited amount of
	 * time. Such surfaces are useful to display things like volume,
	 * brightness and status changes.
	 */
	void (*set_overlay)(struct wl_client *client,
			    struct wl_resource *resource,
			    struct wl_resource *surface);
	/**
	 * set_lock_surface - set lock role
	 * @surface: (none)
	 *
	 * The surface set by this request will receive the lock role.
	 * The lock surface is shown by the compositor when the session is
	 * locked, users interact with it to unlock the session. Only one
	 * surface can have the lock role.
	 */
	void (*set_lock_surface)(struct wl_client *client,
				 struct wl_resource *resource,
				 struct wl_resource *surface);
	/**
	 * set_grab_surface - set grab surface
	 * @surface: (none)
	 *
	 * The surface set by this request will receive a fake
	 * pointer.enter event during grabs at position 0, 0 and is
	 * expected to set an appropriate cursor image as described by the
	 * grab_cursor event sent just before the enter event.
	 */
	void (*set_grab_surface)(struct wl_client *client,
				 struct wl_resource *resource,
				 struct wl_resource *surface);
	/**
	 * set_position - change surface position
	 * @surface: (none)
	 * @x: (none)
	 * @y: (none)
	 *
	 * Move the surface to new coordinates.
	 */
	void (*set_position)(struct wl_client *client,
			     struct wl_resource *resource,
			     struct wl_resource *surface,
			     int32_t x,
			     int32_t y);
	/**
	 * quit - quit the compositor
	 *
	 * Tell the server to quit. The shell client sends this request
	 * after the session has been closed to quit the compositor.
	 */
	void (*quit)(struct wl_client *client,
		     struct wl_resource *resource);
	/**
	 * lock - lock the session
	 *
	 * Tell the server to lock the session.
	 */
	void (*lock)(struct wl_client *client,
		     struct wl_resource *resource);
	/**
	 * unlock - unlock the session
	 *
	 * Tell the server to unlock the session.
	 */
	void (*unlock)(struct wl_client *client,
		       struct wl_resource *resource);
	/**
	 * desktop_ready - desktop is ready to be shown
	 *
	 * Tell the server, that enough desktop elements have been drawn
	 * to make the desktop look ready for use. During start-up, the
	 * server can wait for this request with a black screen before
	 * starting to fade in the desktop, for instance. If the client
	 * parts of a desktop take a long time to initialize, we avoid
	 * showing temporary garbage.
	 */
	void (*desktop_ready)(struct wl_client *client,
			      struct wl_resource *resource);
	/**
	 * minimize_windows - minimize all windows
	 *
	 * Tell the server to minimize all windows.
	 */
	void (*minimize_windows)(struct wl_client *client,
				 struct wl_resource *resource);
	/**
	 * restore_windows - restore all windows
	 *
	 * Tell the server to restore all windows.
	 */
	void (*restore_windows)(struct wl_client *client,
				struct wl_resource *resource);
	/**
	 * add_workspace - add workspace
	 *
	 * Tell the server to add a new workspace.
	 */
	void (*add_workspace)(struct wl_client *client,
			      struct wl_resource *resource);
	/**
	 * select_workspace - select workspace
	 * @workspace: (none)
	 *
	 * Tell the server to select the specified workspace.
	 */
	void (*select_workspace)(struct wl_client *client,
				 struct wl_resource *resource,
				 struct wl_resource *workspace);
};

#define HAWAII_SHELL_LOADED	0
#define HAWAII_SHELL_CONFIGURE	1
#define HAWAII_SHELL_PREPARE_LOCK_SURFACE	2
#define HAWAII_SHELL_GRAB_CURSOR	3
#define HAWAII_SHELL_WINDOW_MAPPED	4
#define HAWAII_SHELL_WINDOW_SWITCHING_STARTED	5
#define HAWAII_SHELL_WINDOW_SWITCHING_FINISHED	6
#define HAWAII_SHELL_WINDOW_SWITCHED	7
#define HAWAII_SHELL_WORKSPACE_ADDED	8

#define HAWAII_SHELL_LOADED_SINCE_VERSION	1
#define HAWAII_SHELL_CONFIGURE_SINCE_VERSION	1
#define HAWAII_SHELL_PREPARE_LOCK_SURFACE_SINCE_VERSION	1
#define HAWAII_SHELL_GRAB_CURSOR_SINCE_VERSION	1
#define HAWAII_SHELL_WINDOW_MAPPED_SINCE_VERSION	1
#define HAWAII_SHELL_WINDOW_SWITCHING_STARTED_SINCE_VERSION	1
#define HAWAII_SHELL_WINDOW_SWITCHING_FINISHED_SINCE_VERSION	1
#define HAWAII_SHELL_WINDOW_SWITCHED_SINCE_VERSION	1
#define HAWAII_SHELL_WORKSPACE_ADDED_SINCE_VERSION	1

static inline void
hawaii_shell_send_loaded(struct wl_resource *resource_)
{
	wl_resource_post_event(resource_, HAWAII_SHELL_LOADED);
}

static inline void
hawaii_shell_send_configure(struct wl_resource *resource_, struct wl_resource *surface, int32_t width, int32_t height)
{
	wl_resource_post_event(resource_, HAWAII_SHELL_CONFIGURE, surface, width, height);
}

static inline void
hawaii_shell_send_prepare_lock_surface(struct wl_resource *resource_)
{
	wl_resource_post_event(resource_, HAWAII_SHELL_PREPARE_LOCK_SURFACE);
}

static inline void
hawaii_shell_send_grab_cursor(struct wl_resource *resource_, uint32_t cursor)
{
	wl_resource_post_event(resource_, HAWAII_SHELL_GRAB_CURSOR, cursor);
}

static inline void
hawaii_shell_send_window_mapped(struct wl_resource *resource_, struct wl_resource *id, const char *title, const char *identifier, int32_t state)
{
	wl_resource_post_event(resource_, HAWAII_SHELL_WINDOW_MAPPED, id, title, identifier, state);
}

static inline void
hawaii_shell_send_window_switching_started(struct wl_resource *resource_)
{
	wl_resource_post_event(resource_, HAWAII_SHELL_WINDOW_SWITCHING_STARTED);
}

static inline void
hawaii_shell_send_window_switching_finished(struct wl_resource *resource_)
{
	wl_resource_post_event(resource_, HAWAII_SHELL_WINDOW_SWITCHING_FINISHED);
}

static inline void
hawaii_shell_send_window_switched(struct wl_resource *resource_, struct wl_resource *window)
{
	wl_resource_post_event(resource_, HAWAII_SHELL_WINDOW_SWITCHED, window);
}

static inline void
hawaii_shell_send_workspace_added(struct wl_resource *resource_, struct wl_resource *workspace, int32_t active)
{
	wl_resource_post_event(resource_, HAWAII_SHELL_WORKSPACE_ADDED, workspace, active);
}

/**
 * hawaii_panel_manager - create panels
 * @set_panel: set panel role
 *
 * This interface is used by the shell to create panels. Only one client
 * can bind.
 */
struct hawaii_panel_manager_interface {
	/**
	 * set_panel - set panel role
	 * @id: (none)
	 * @surface: (none)
	 *
	 * The surface set by this request will receive the panel role.
	 * Panels display launchers, buttons and other kind of elements.
	 */
	void (*set_panel)(struct wl_client *client,
			  struct wl_resource *resource,
			  uint32_t id,
			  struct wl_resource *surface);
};


#ifndef HAWAII_PANEL_EDGE_ENUM
#define HAWAII_PANEL_EDGE_ENUM
enum hawaii_panel_edge {
	HAWAII_PANEL_EDGE_LEFT = 0,
	HAWAII_PANEL_EDGE_TOP = 1,
	HAWAII_PANEL_EDGE_RIGHT = 2,
	HAWAII_PANEL_EDGE_BOTTOM = 3,
};
#endif /* HAWAII_PANEL_EDGE_ENUM */

#ifndef HAWAII_PANEL_ALIGNMENT_ENUM
#define HAWAII_PANEL_ALIGNMENT_ENUM
enum hawaii_panel_alignment {
	HAWAII_PANEL_ALIGNMENT_LEFT = 0,
	HAWAII_PANEL_ALIGNMENT_CENTER = 1,
	HAWAII_PANEL_ALIGNMENT_RIGHT = 2,
};
#endif /* HAWAII_PANEL_ALIGNMENT_ENUM */

struct hawaii_panel_interface {
	/**
	 * set_alignment - set panel alignment
	 * @alignment: (none)
	 *
	 * Set panel alignment.
	 */
	void (*set_alignment)(struct wl_client *client,
			      struct wl_resource *resource,
			      uint32_t alignment);
	/**
	 * set_offset - set panel offset
	 * @offset: (none)
	 *
	 * Set panel offset.
	 */
	void (*set_offset)(struct wl_client *client,
			   struct wl_resource *resource,
			   uint32_t offset);
	/**
	 * set_thickness - set panel thickness
	 * @thickness: (none)
	 *
	 * Set panel size.
	 */
	void (*set_thickness)(struct wl_client *client,
			      struct wl_resource *resource,
			      uint32_t thickness);
	/**
	 * set_length - set panel length
	 * @length: (none)
	 *
	 * Set panel length.
	 */
	void (*set_length)(struct wl_client *client,
			   struct wl_resource *resource,
			   uint32_t length);
	/**
	 * set_min_length - set minimum length
	 * @min_length: (none)
	 *
	 * Set minimum length of this panel.
	 */
	void (*set_min_length)(struct wl_client *client,
			       struct wl_resource *resource,
			       uint32_t min_length);
	/**
	 * set_max_length - set maximum length
	 * @max_length: (none)
	 *
	 * Set maximum length of this panel.
	 */
	void (*set_max_length)(struct wl_client *client,
			       struct wl_resource *resource,
			       uint32_t max_length);
	/**
	 * dock - dock the panel
	 * @edge: (none)
	 * @output: (none)
	 *
	 * Set panel position on the specified output according to edge,
	 * alignment and offset.
	 */
	void (*dock)(struct wl_client *client,
		     struct wl_resource *resource,
		     uint32_t edge,
		     struct wl_resource *output);
};

#define HAWAII_PANEL_DOCKED	0

#define HAWAII_PANEL_DOCKED_SINCE_VERSION	1

static inline void
hawaii_panel_send_docked(struct wl_resource *resource_)
{
	wl_resource_post_event(resource_, HAWAII_PANEL_DOCKED);
}

struct hawaii_shell_surface_interface {
	/**
	 * set_popup - make the surface a popup surface
	 * @id: (none)
	 * @parent: (none)
	 * @surface: (none)
	 * @x: (none)
	 * @y: (none)
	 *
	 * Map the surface as a popup.
	 *
	 * A popup surface is a surface with an added pointer grab.
	 *
	 * An existing implicit grab will be changed to owner-events mode,
	 * and the popup grab will continue after the implicit grab ends
	 * (i.e. releasing the mouse button does not cause the popup to be
	 * unmapped).
	 *
	 * The popup grab continues until the window is destroyed or a
	 * mouse button is pressed in any other window.
	 */
	void (*set_popup)(struct wl_client *client,
			  struct wl_resource *resource,
			  uint32_t id,
			  struct wl_resource *parent,
			  struct wl_resource *surface,
			  int32_t x,
			  int32_t y);
	/**
	 * set_dialog - set dialog role
	 * @output: (none)
	 * @surface: (none)
	 *
	 * The surface set by this request will receive the dialog role.
	 * Dialogs have a modal behavior which interrupts user's workflow.
	 */
	void (*set_dialog)(struct wl_client *client,
			   struct wl_resource *resource,
			   struct wl_resource *output,
			   struct wl_resource *surface);
};


#define HAWAII_POPUP_SURFACE_POPUP_DONE	0

#define HAWAII_POPUP_SURFACE_POPUP_DONE_SINCE_VERSION	1

static inline void
hawaii_popup_surface_send_popup_done(struct wl_resource *resource_)
{
	wl_resource_post_event(resource_, HAWAII_POPUP_SURFACE_POPUP_DONE);
}

#define HAWAII_KEY_BINDING_TRIGGERED	0

#define HAWAII_KEY_BINDING_TRIGGERED_SINCE_VERSION	1

static inline void
hawaii_key_binding_send_triggered(struct wl_resource *resource_)
{
	wl_resource_post_event(resource_, HAWAII_KEY_BINDING_TRIGGERED);
}

/**
 * hawaii_window - interface to control client windows
 * @set_state: set client window state
 * @close: close window
 *
 * Only one client can bind this interface at a time.
 */
struct hawaii_window_interface {
	/**
	 * set_state - set client window state
	 * @state: (none)
	 *
	 * Set client window state.
	 */
	void (*set_state)(struct wl_client *client,
			  struct wl_resource *resource,
			  int32_t state);
	/**
	 * close - close window
	 *
	 * Close the client window.
	 */
	void (*close)(struct wl_client *client,
		      struct wl_resource *resource);
};

#define HAWAII_WINDOW_TITLE_CHANGED	0
#define HAWAII_WINDOW_IDENTIFIER_CHANGED	1
#define HAWAII_WINDOW_STATE_CHANGED	2
#define HAWAII_WINDOW_UNMAPPED	3

#define HAWAII_WINDOW_TITLE_CHANGED_SINCE_VERSION	1
#define HAWAII_WINDOW_IDENTIFIER_CHANGED_SINCE_VERSION	1
#define HAWAII_WINDOW_STATE_CHANGED_SINCE_VERSION	1
#define HAWAII_WINDOW_UNMAPPED_SINCE_VERSION	1

static inline void
hawaii_window_send_title_changed(struct wl_resource *resource_, const char *title)
{
	wl_resource_post_event(resource_, HAWAII_WINDOW_TITLE_CHANGED, title);
}

static inline void
hawaii_window_send_identifier_changed(struct wl_resource *resource_, const char *identifier)
{
	wl_resource_post_event(resource_, HAWAII_WINDOW_IDENTIFIER_CHANGED, identifier);
}

static inline void
hawaii_window_send_state_changed(struct wl_resource *resource_, int32_t state)
{
	wl_resource_post_event(resource_, HAWAII_WINDOW_STATE_CHANGED, state);
}

static inline void
hawaii_window_send_unmapped(struct wl_resource *resource_)
{
	wl_resource_post_event(resource_, HAWAII_WINDOW_UNMAPPED);
}

/**
 * hawaii_workspace - control and manager workspaces
 * @remove: remove workspace
 *
 * Each workspace can bind this interface.
 */
struct hawaii_workspace_interface {
	/**
	 * remove - remove workspace
	 *
	 * Remove this workspace.
	 */
	void (*remove)(struct wl_client *client,
		       struct wl_resource *resource);
};

#define HAWAII_WORKSPACE_ACTIVATED	0
#define HAWAII_WORKSPACE_DEACTIVATED	1

#define HAWAII_WORKSPACE_ACTIVATED_SINCE_VERSION	1
#define HAWAII_WORKSPACE_DEACTIVATED_SINCE_VERSION	1

static inline void
hawaii_workspace_send_activated(struct wl_resource *resource_)
{
	wl_resource_post_event(resource_, HAWAII_WORKSPACE_ACTIVATED);
}

static inline void
hawaii_workspace_send_deactivated(struct wl_resource *resource_)
{
	wl_resource_post_event(resource_, HAWAII_WORKSPACE_DEACTIVATED);
}

#ifdef  __cplusplus
}
#endif

#endif
